#!/usr/bin/env python

import datetime
import sys
from gettext import bindtextdomain, textdomain, gettext as _

import gtk
import gnomeapplet
import gobject
import pango

### Running in panel applet mode.
import gimmie_globals
gimmie_globals.gimmie_is_panel_applet = lambda: True

from gimmie_base import Topic, ItemSource
from gimmie_applications import ApplicationsTopic
from gimmie_computer import ComputerTopic
from gimmie_documents import DocumentsTopic
from gimmie_people import PeopleTopic
from gimmie_topicwin import TopicWindowBase
from gimmie_util import icon_factory, KillFocusPadding


def color_average(col1, col2, colormap):
    if not col1 or not col2:
        return col1 or col2
    new = gtk.gdk.Color()
    new.red = int((col1.red + col2.red) / 2)
    new.green = int((col1.green + col2.green) / 2)
    new.blue = int((col1.blue + col2.blue) / 2)
    return colormap.alloc_color(new)


def color_average_widget_state(col1, widget, state):
    state_col = widget.get_style().bg[state]
    return color_average(col1, state_col, widget.get_colormap())


class TimeBar(gtk.EventBox):
    __gsignals__ = {
        "zoom-changed" : (gobject.SIGNAL_RUN_LAST,
                          gobject.TYPE_NONE,
                          (gobject.TYPE_INT,)),
        }

    _zoomlist = [[_("Today"), 1],
                 [_("Last 3 days"), 3],
                 [_("Last week"), 7],
                 [_("Last 2 weeks"), 14],
                 [_("Last month"), 30],
                 [_("Last 3 months"), 90],
                 [_("Forever"), -1]]
    _default_zoom = 7

    def __init__(self):
        gtk.EventBox.__init__(self)
        self.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse("white"))
        
        box = gtk.HBox(False, 0)
        box.show()
        self.add(box)

        spacer = gtk.HSeparator()
        box.pack_start(spacer, True, True, 6)

        self.label = gtk.Label("Last 2 weeks")
        box.pack_start(self.label, False, False, 6)

        self.in_btn = gtk.Button()
        self.in_btn.add(gtk.image_new_from_stock(gtk.STOCK_REMOVE, gtk.ICON_SIZE_SMALL_TOOLBAR))
        self.in_btn.set_relief(gtk.RELIEF_NONE)
        self.in_btn.connect("clicked", lambda btn: self._zoom_clicked(-1))
        box.pack_start(self.in_btn, False, False)

        self.out_btn = gtk.Button()
        self.out_btn.add(gtk.image_new_from_stock(gtk.STOCK_ADD, gtk.ICON_SIZE_SMALL_TOOLBAR))
        self.out_btn.set_relief(gtk.RELIEF_NONE)
        self.out_btn.connect("clicked", lambda btn: self._zoom_clicked(1))
        box.pack_start(self.out_btn, False, False)

        box.show_all()

        self.zoom_level = self._default_zoom

    def get_zoom_level(self):
        return self.zoom_level

    def get_zoom_level_list(self, floor = _default_zoom, include_forever = True):
        return [days for name, days in self._zoomlist \
                if days >= floor or (days < 0 and include_forever)]

    def _zoom_clicked(self, dir):
        for i in range(len(self._zoomlist)):
            label, days = self._zoomlist[i]
            if days == self.zoom_level:
                try:
                    label, days = self._zoomlist[i + dir]
                    self.set_zoom_level(days, label)
                    break
                except IndexError:
                    pass

    def set_zoom_level(self, zoom_level, label=None):
        self.zoom_level = zoom_level

        if not label:
            for _label, days in self._zoomlist:
                if days == zoom_level:
                    label = _label
        self.label.set_text(label)
        self.emit("zoom-changed", zoom_level)


class GrabOnShowWindow(gtk.Window):
    __gsignals__ = {
        'grab-broken-event': 'override',
        'map-event' : 'override',
        'unmap-event' : 'override',
        'button-press-event' : 'override',
        'key-press-event' : 'override',
        }

    def __init__(self):
        gtk.Window.__init__(self)
        self.grabbed = False

    def _grab_window_event(self, win, ev):
        if ev.type in (gtk.gdk.UNMAP, gtk.gdk.SELECTION_CLEAR):
            self.grabbed = False # Force a regrab
            self.do_map_event(ev) 
        return False

    def do_grab_broken_event(self, ev):
        if ev.grab_window:
            ev_widget = ev.grab_window.get_user_data()
            if isinstance(ev_widget, gtk.Widget):
                ev_widget.connect("event", self._grab_window_event)
        return True

    def do_map_event(self, ev):
        if not self.grabbed:
            self.grab_focus()
            self.grab_add()

            time = gtk.get_current_event_time()
            gtk.gdk.pointer_grab(self.window, True, gtk.gdk.BUTTON_PRESS_MASK, None, None, time)
            gtk.gdk.keyboard_grab(self.window, True, time)
        self.grabbed = True

    def do_unmap_event(self, ev):
        if self.grabbed:
            time = gtk.get_current_event_time()
            gtk.gdk.pointer_ungrab(time)
            gtk.gdk.keyboard_ungrab(time)
            self.grab_remove()
        self.grabbed = False

    def do_button_press_event(self, ev):
        win_tuple = gtk.gdk.window_at_pointer()
        ev_win = (win_tuple and win_tuple[0]) or None

        if not ev_win:
            # External application, hide and give up grab
            self.hide()
        elif ev_win.get_toplevel() != self.window:
            # Other toplevel window, hide and forward the event
            self.hide()
            ev_widget = ev_win.get_user_data()
            return ev_widget.do_button_press_event(ev)

        return True

    def do_key_press_event(self, ev):
        if not gtk.Window.do_key_press_event(self, ev) \
               and ev.keyval == gtk.gdk.keyval_from_name("Escape"):
            self.hide()
            return True


class TopicWindowMod(GrabOnShowWindow, TopicWindowBase):
    '''
    Panel applet menu version of topic window.
    '''
    def __init__(self, topic):
        GrabOnShowWindow.__init__(self)
        TopicWindowBase.__init__(self, topic)

        self.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_MENU)
        self.set_decorated(False)
        self.set_default_size(660, -1)

        self.add_accel_group(self.accel_group)

        ### Merge the topic's color with the style as a border
        new = color_average_widget_state(topic.get_hint_color(), self, gtk.STATE_NORMAL)
        self.modify_bg(gtk.STATE_NORMAL, new)

        frame = gtk.Frame()
        frame.show()
        frame.set_shadow_type(gtk.SHADOW_OUT)
        self.add(frame)

        ev_box = gtk.EventBox()
        ev_box.show()
        frame.add(ev_box)

        ### Merge the topic's color with the style as a border
        new = color_average_widget_state(topic.get_hint_color(), ev_box, gtk.STATE_ACTIVE)
        ev_box.modify_bg(gtk.STATE_NORMAL, new)

        ev_box2 = gtk.EventBox()
        ev_box2.show()
        ev_box2.set_border_width(4)
        ev_box.add(ev_box2)
        ev_box = ev_box2

        frame = gtk.Frame()
        frame.show()
        frame.set_shadow_type(gtk.SHADOW_OUT)
        ev_box.add(frame)

        # Contains the visual frame, giving it some space
        content = gtk.VBox(False, 0)
        content.set_border_width(0)
        content.show()
        frame.add(content)

        # Hbox containing the sidebar buttons and the toolbar/iconview
        body = gtk.HBox(False, 12)
        body.set_border_width(12)
        body.show()
        content.pack_start(body, True, True, 0)

        # Toolbar
        content.pack_start(self.toolbar, False, False, 0)

        # Load up the sidebar
        body.pack_start(self.sidebar, False, False, 0)

        frame = gtk.Frame()
        frame.show()
        frame.set_shadow_type(gtk.SHADOW_IN)
        body.pack_start(frame, True, True, 0)

        # Vbox containing the toolbar and iconview
        vbox = gtk.VBox(False, 0)
        vbox.show()
        frame.add(vbox)

        self.timebar = TimeBar()
        self.timebar.connect("zoom-changed", self._zoom_changed)
        vbox.pack_start(self.timebar, False, False)

        # Scrolled window containing the icon view
        vbox.pack_start(self.scroll, True, True)

        # Iconview for the current sidebar selection
        self.view.connect("item-activated", lambda v, p: self.hide()) # Iconify on item open

        # Setup the toolbar
        self._add_toolbar_items()

        # Select an initial sidebar button
        self._find_first_button()

    def _source_set(self, sidebar, source):
        print " *** Setting source:", source

        self.active_source = source
        if source:
            # Clear any search text
            self.search_tool_item.cancel()

            # FIXME: This loses the focused item
            if source.get_filter_by_date():
                filtered = self._filter_items_num_days_ago(source.get_items(),
                                                           self.timebar.get_zoom_level())
                self.view.load_items(filtered, lambda: self.timebar.show())
            else:
                self.view.load_items(source.get_items(), lambda: self.timebar.hide())
        else:
            self.timebar.hide()
            self.view.get_model().clear()

    def _find_first_button(self):
        '''
        Activates the first sidebar button which is not insensitive, and sets
        the zoom level (if possible) so at least two items are displayed.
        FIXME: Populate with default items if first button is empty?
        '''
        if self.active_source:
            return

        min_items = 5

        for source in self.topic.get_sidebar_source_list():
            if isinstance(source, ItemSource) and source.get_enabled():
                items = source.get_items()
                if len(items) < min_items:
                    continue # Skip to the next sidebar button

                if source.get_filter_by_date():
                    for zoom in self.timebar.get_zoom_level_list():
                        filtered = self._filter_items_num_days_ago(items, zoom)
                        if len(filtered) >= min_items:
                            self.timebar.set_zoom_level(zoom)
                            break
                    else:
                        continue # Skip to the next sidebar button

                # FIXME: We're filtering to get an item count, and then
                # refiltering to show the items in _source_set.
                self.sidebar.set_source(source)
                break

        # When run inside an idle handler, don't run again
        return False

    def _make_toolbar_expander(self):
        sep = gtk.SeparatorToolItem()
        sep.set_draw(False)
        sep.set_expand(True)
        sep.show()
        return sep

    def _add_toolbar_items(self):
        i = gtk.ToolButton(icon_factory.load_image("gnome-favorites", gtk.ICON_SIZE_LARGE_TOOLBAR))
        i.set_sensitive(False)
        i.set_tooltip(self.tooltips, "Add to Favorites")
        i.show_all()
        self.toolbar.insert(i, -1)

        sep = gtk.SeparatorToolItem()
        sep.set_draw(False)
        sep.show()
        self.toolbar.insert(sep, -1)

        for i in self.topic.get_toolbar_items(self.tooltips):
            #self.toolbar.insert(self._make_toolbar_expander(), -1)
            if i:
                i.show_all()
                self.toolbar.insert(i, -1)

        # Right-align the zoom and search tool items
        self.toolbar.insert(self._make_toolbar_expander(), -1)

        ### Disabled in favor of timebar
        #self._add_zoom_menu()
        
        self._add_search()


def get_topic_window_mod(self):
    if not self.topic_window:
        self.topic_window = TopicWindowMod(self)
    return self.topic_window

### Override the Topic window to use the appler menu window class.
Topic.get_topic_window = get_topic_window_mod


class TopicButtonMod(gtk.ToggleButton):
    __gsignals__ = {
        'button-press-event' : 'override',
        'activate' : 'override'
        }

    def __init__(self, topic, edge_gravity, icon):
        gtk.Button.__init__(self)
        self.set_property("can-default", False)
        self.set_property("can-focus", False)
        self.set_border_width(0)
        
        KillFocusPadding(self, "topic-button")

        ### Don't use any border, but loses background
        #self.set_relief(gtk.RELIEF_NONE)
        
        self.set_image(icon)
        self.set_label(topic.get_name())
                
        self.topic = topic
        self.topic_win = None

    def do_realize(self):
        gtk.ToggleButton.do_realize(self)

        color = self.topic.get_hint_color()

        new = color_average_widget_state(color, self, gtk.STATE_NORMAL)
        self.modify_bg(gtk.STATE_NORMAL, new)

        new = color_average_widget_state(color, self, gtk.STATE_ACTIVE)
        self.modify_bg(gtk.STATE_ACTIVE, new)

        new = color_average_widget_state(color, self, gtk.STATE_PRELIGHT)
        self.modify_bg(gtk.STATE_PRELIGHT, new)

    def do_button_press_event(self, ev):
        if ev.button == 3:
            # FIXME: Add items to panel right-click menu
            return False
        
            menu = gtk.Menu()

            for mi in self.topic.get_context_menu_items():
                if not mi:
                    mi = gtk.SeparatorMenuItem()
                    mi.show()
                menu.append(mi)

            menu.connect('selection-done', lambda x: menu.destroy())

            menu.attach_to_widget(self, None)
            menu.popup(None, None, None, ev.button, ev.time)

            return True
        else:
            if not self.get_active():
                self._show()
            return True

    def _show(self):
        if not self.topic_win:
            self.topic_win = self.topic.get_topic_window()
            self.topic_win.realize()
            #self.do_set_wm_icon_geometry()

            self.topic_win.connect("map-event", self._map)
            self.topic_win.connect("unmap-event", self._unmap)

        self._position(self.topic_win)
        self.topic_win.present()

    def _map(self, win, ev):
        self.set_active(True)

    def _unmap(self, win, ev):
        self.set_active(False)

    def _position(self, win):
        x, y = self.window.get_origin()

        alloc = self.allocation
        x += alloc.x + 1
        y += alloc.y - 1

        width, height = win.get_size_request()
        if y + height >= self.get_screen().get_height:
            y -= height
        else:
            y += alloc.height

        win.move(x, y)


def gimmie_applet_factory(applet, iid):
    swapbar = 0
    autohide = 0
    vertical = 0

    edge_gravity = (vertical and gtk.gdk.GRAVITY_WEST) or gtk.gdk.GRAVITY_SOUTH

    content = gtk.HBox(True, 0)
    content.show()
    applet.add(content)
    applet.show()

    topic_and_icon = [ (ComputerTopic(), "computer"),
                       (ApplicationsTopic(), "gnome-main-menu"),
                       (DocumentsTopic(), "stock_new-template"),
                       (PeopleTopic(), "stock_people") ]
    for topic, icon in topic_and_icon:
        btn = TopicButtonMod(topic, edge_gravity, icon_factory.load_image(icon, 18))
        btn.show()
        content.pack_start(btn, True, True, 0)

    KillFocusPadding(applet, "applet")
    
    return True


def main(args):
    bindtextdomain('gimmie', gimmie_globals.localedir)
    textdomain('gimmie')

    if len(sys.argv) == 2 and sys.argv[1] == "window":
        main_window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        main_window.set_title("Gimmie")
        main_window.connect("destroy", gtk.main_quit)
        app = gnomeapplet.Applet()
        gimmie_applet_factory(app, None)
        app.reparent(main_window)
        main_window.show_all()
        gtk.main()
        sys.exit()

    gnomeapplet.bonobo_factory("OAFIID:GNOME_Gimmie_Factory", 
                               gnomeapplet.Applet.__gtype__, 
                               "gimmie", "0", gimmie_applet_factory)


if __name__ == "__main__":
    main(None)
